from __future__ import with_statement
import sys
import unittest2
from cStringIO import StringIO as SIO

import support

from chemfp import io
from chemfp import ParseError

# Make sure the module correctly implements __all__
before = after = None
before = set(globals())
from chemfp.sdf_reader import *
after = set(globals())
assert len(after - before) == 4, ("wrong import * count", after-before)


TRYPTOPHAN_SDF = support.fullpath("tryptophan.sdf")
PUBCHEM_SDF = support.fullpath("pubchem.sdf")
PUBCHEM_SDF_GZ = support.fullpath("pubchem.sdf.gz")
STRANGE_SDF = support.fullpath("strange.sdf")

expected_identifiers = ["9425004", "9425009", "9425012", "9425015",
                        "9425018", "9425021", "9425030", "9425031",
                        "9425032", "9425033", "9425034", "9425035",
                        "9425036", "9425037", "9425040", "9425041",
                        "9425042", "9425045", "9425046"]
expected_linenos = [1, 191, 401, 592, 817, 1027, 1236, 1457, 1678,
                    1872, 2086, 2288, 2493, 2700, 2894, 3103, 3305,
                    3507, 3722]
assert len(expected_identifiers) == len(expected_linenos)
expected_locs = [dict(title=title, lineno=lineno) for (title, lineno) in
                     zip(expected_identifiers, expected_linenos)]

class TestReadRecords(unittest2.TestCase):
    def test_reads_the_only_record(self):
        n = sum(1 for x in open_sdf(TRYPTOPHAN_SDF))
        self.assertEquals(n, 1)
    def test_reads_all_records(self):
        n = sum(1 for x in open_sdf(PUBCHEM_SDF))
        self.assertEquals(n, 19)
    def test_reads_all_compressed_records(self):
        n = sum(1 for x in open_sdf(PUBCHEM_SDF_GZ))
        self.assertEquals(n, 19)

    def test_reads_from_stdin(self):
        old_stdin = sys.stdin
        sys.stdin = open(PUBCHEM_SDF, "rb")
        try:
            n = sum(1 for x in open_sdf())
        finally:
            sys.stdin = sys.stdin
        self.assertEquals(n, 19)

    def test_reads_from_gzip_stdin(self):
        old_stdin = sys.stdin
        sys.stdin = open(PUBCHEM_SDF_GZ, "rb")
        try:
            n = sum(1 for x in open_sdf(None, "gzip"))
        finally:
            sys.stdin = sys.stdin
        self.assertEquals(n, 19)

    def test_reads_from_fileobj(self):
        f = open(PUBCHEM_SDF, "rU")
        n = sum(1 for x in open_sdf(f))
        self.assertEquals(n, 19)
        
    def test_reads_from_uncompressed_fileobj(self):
        f = open(PUBCHEM_SDF, "rU")
        n = sum(1 for x in open_sdf(f, "none"))
        self.assertEquals(n, 19)

    def test_reads_from_gzip_fileobj(self):
        f = open(PUBCHEM_SDF_GZ, "rb")
        n = sum(1 for x in open_sdf(f, "gzip"))
        self.assertEquals(n, 19)

    def test_handles_loc(self):
        loc = io.Location.from_source(PUBCHEM_SDF_GZ)
        self.assertEquals(loc.filename, PUBCHEM_SDF_GZ)
        
        results = []
        for x in open_sdf(PUBCHEM_SDF_GZ, location=loc):
            results.append(dict(title=loc.first_line,
                                lineno=loc.lineno))
        self.assertEquals(results, expected_locs)

    def test_when_using_wrong_compression(self):
        try:
            n = sum(1 for x in open_sdf(PUBCHEM_SDF, "gzip"))
            raise AssertionError("parsed a gzip'ed file?")
        except IOError:
            pass
            
        

tryptophan = open(TRYPTOPHAN_SDF).read()

class ReadReturnsSmallAmounts(object):
    def __init__(self):
        self.f = open(PUBCHEM_SDF, "rU")
    def read(self, n):
        return self.f.read(1)

class ReadReturnsOneRecord(object):
    def __init__(self):
        self.f = open_sdf(PUBCHEM_SDF)
    def read(self, n):
        try:
            rec = self.f.next()
        except StopIteration:
            return ""
        assert rec.endswith("\n$$$$\n")
        rec = rec[:-6] + ">spam\n" + ("X" * (n-len(rec)-6)) + "\n$$$$\n"
        assert len(rec) == n
        return rec

class ReadReturnsTwoRecords(object):
    def __init__(self):
        self.f = open_sdf(PUBCHEM_SDF)
    def read(self, n):
        try:
            rec = self.f.next()
        except StopIteration:
            return ""
        try:
            rec2 = self.f.next()
        except StopIteration:
            return rec
        rec = rec + rec2
        assert rec.endswith("\n$$$$\n")
        rec = rec[:-6] + ">spam\n" + ("X" * (n-len(rec)-6)) + "\n$$$$\n"
        assert len(rec) == n
        return rec

class TestBoundaryConditions(unittest2.TestCase):
    def test_missing_terminal_newline(self):
        f = SIO(tryptophan.rstrip("\n"))
        n = sum(1 for x in open_sdf(f))
        self.assertEquals(n, 1)
    def test_small_amounts(self):
        # the O(n**2) behavior really hits hard here - this takes a full second to work
        n = sum(1 for x in open_sdf(ReadReturnsSmallAmounts()))
        self.assertEquals(n, 19)
    def test_exact_record_boundary_reads(self):
        loc = io.Location()
        titles = [loc.first_line for x in open_sdf(ReadReturnsOneRecord(), location=loc)]
        self.assertEquals(titles, expected_identifiers)
    def test_two_record_boundary_reads(self):
        loc = io.Location()
        titles = [loc.first_line for x in open_sdf(ReadReturnsTwoRecords(), location=loc)]
        self.assertEquals(titles, expected_identifiers)

class TestReadErrors(unittest2.TestCase):
    def test_wrong_format(self):
        f = SIO("Spam\n")
        try:
            for x in open_sdf(f):
                raise AssertionError("Bad parse")
        except ParseError, err:
            self.assertIn("Could not find a valid SD record", str(err))
            self.assertIn("line 1", str(err), True)
            
    def test_record_too_large(self):
        f = SIO( (tryptophan * ((2000000 // len(tryptophan)) + 1)).replace("$$$$", "1234"))
        try:
            for x in open_sdf(f):
                raise AssertionError("should not be able to read the first record")
        except ParseError, err:
            self.assertIn("too large", str(err))
            self.assertIn(", line 1", str(err))

    def test_has_extra_data(self):
        f = SIO(tryptophan + tryptophan + "blah")
        try:
            for i, x in enumerate(open_sdf(f)):
                if i > 1:
                    raise AssertionError("bad record count")
        except ParseError, err:
            self.assertIn("unexpected content", str(err))
            expected_lineno = (tryptophan.count("\n")*2) + 1
            expected_lineno_msg = ", line %d" % expected_lineno
            self.assertIn(expected_lineno_msg, str(err))

    def test_bad_format(self):
        f = SIO(tryptophan + tryptophan.replace("V2000", "V4000"))
        try:
            for i, x in enumerate(open_sdf(f)):
                if i > 0:
                    raise AssertionError("bad record count")
        except ParseError as err:
            self.assertIn("incorrectly formatted record", str(err))
            self.assertIn(", line 70", str(err))

    def test_my_error_handler(self):
        class CaptureErrors(object):
            def __init__(self):
                self.errors = []
            def error(self, msg, loc):
                self.errors.append( (msg, {"lineno": loc.lineno,
                                           "title": loc.first_line}) )

        my_error_handler = CaptureErrors()
        loc = io.Location()
        f = SIO(tryptophan + tryptophan.replace("V2000", "V4000") + tryptophan)
        titles = [loc.lineno for rec in open_sdf(f, location=loc,
                                                    errors=my_error_handler)]
        self.assertEquals(titles, [1, 137])
        self.assertEquals(my_error_handler.errors, [("incorrectly formatted record",
                                                     {"lineno": 70,
                                                      "title": "tryptophan.pdb"})])
        

expected_hbond_donors = ["2","2","2","2","2","2","4","4","2", "2",
                         "2","2","2","3","2","3","3","2","2"]
expected_complexity = ["491", "513", "419", "597", "545", "590", "660", "660",
                       "394", "544", "458", "589", "532", "506", "640", "557",
                       "557", "520", "437"]
expected_xlogp = ["2.8", "1.9", "1", "3.3", "1.5", "2.6", None, "-0.9",
                  "2", "2.1", "2.9", "1.7", "-1.5", "0.4", "0.6",
                  "0.4", "0.4", "2", "2.5"]

assert len(expected_complexity) == len(expected_hbond_donors) == len(expected_linenos)
assert len(expected_xlogp) == len(expected_linenos)

class TestIterTwoTags(unittest2.TestCase):
    def test_read_two_existing_tags(self):
        fields = list(iter_two_tags(open_sdf(PUBCHEM_SDF),
                            "PUBCHEM_CACTVS_HBOND_DONOR", "PUBCHEM_CACTVS_COMPLEXITY"))
        self.assertEquals(fields, zip(expected_hbond_donors, expected_complexity))

    def test_read_tag_missing_data_field1(self):
        fields = list(iter_two_tags(open_sdf(PUBCHEM_SDF),
                                    "PUBCHEM_CACTVS_XLOGP", "PUBCHEM_CACTVS_HBOND_DONOR"))
        self.assertEquals(fields, zip(expected_xlogp, expected_hbond_donors))

    def test_read_tag_missing_data_field2(self):
        fields = list(iter_two_tags(open_sdf(PUBCHEM_SDF),
                                    "PUBCHEM_CACTVS_HBOND_DONOR", "PUBCHEM_CACTVS_XLOGP"))
        self.assertEquals(fields, zip(expected_hbond_donors, expected_xlogp))
        
    def test_edge_conditions1(self):
        fields = list(iter_two_tags(open_sdf(STRANGE_SDF), "noblank", "twolines"))
        self.assertEquals(fields, [("This line is not followed by a blank line",
                                    "This contains two lines"), (None, None)])

    def test_edge_conditions2(self):
        fields = list(iter_two_tags(open_sdf(STRANGE_SDF),
                                    "duplicate", "embedded-tags"))
        self.assertEquals(fields, [("This is the first version.",
                    "I<junk> have<junk> tags  <junk>  in the data line <junk>"),
                                   (None, None)])

    def test_edge_conditions3(self):
        fields = list(iter_two_tags(open_sdf(STRANGE_SDF), "junk", "blank lines"))
        self.assertEquals(fields, [
      ("This line contains some of the strange junk that might exist on the tag line", ""),
            (None, None)])

    def test_edge_conditions4(self):
        fields = list(iter_two_tags(open_sdf(STRANGE_SDF), "nada", "fini"))
        self.assertEquals(fields, [(None, None), ("", "")])

    def test_bad_tags(self):
        for tag in ("<", ">", "\n", "\t", "1<2", "2<1", "blah\t"):
            self.assertRaises(TypeError, iter_two_tags([], tag, "fini"))
            self.assertRaises(TypeError, iter_two_tags([], "fini", tag))
    
class TestReadTitleAndTag(unittest2.TestCase):
    def test_read_existing_tag(self):
        fields = list(iter_title_and_tag(open_sdf(PUBCHEM_SDF),
                                         "PUBCHEM_CACTVS_HBOND_DONOR"))
        self.assertEquals(fields, zip(expected_identifiers, expected_hbond_donors))

    def test_missing_tag(self):
        fields = list(iter_title_and_tag(open_sdf(PUBCHEM_SDF), "PUBCHEM_CACTVS_XLOGP"))
                                         
        self.assertEquals(fields, zip(expected_identifiers, expected_xlogp))

    def test_bad_tags(self):
        for tag in ("<", ">", "\n", "\t", "1<2", "2<1", "blah\t"):
            self.assertRaises(TypeError, iter_title_and_tag([], tag))


if __name__ == "__main__":
    unittest2.main()



import sys
import os
from cStringIO import StringIO
import tempfile
import shutil

import unittest2

try:
    import chemfp_converters
except ImportError:
    has_chemfp_converters = False
else:
    has_chemfp_converters = True

# Ignore the close. io.write_fps1_output() auto-closes its output.
class SIO(object):
    def __init__(self):
        self.sio = StringIO()
    def write(self, s):
        return self.sio.write(s)
    def writelines(self, lines):
        return self.sio.writelines(lines)
    def close(self):
        # Ignore this
        pass
    def flush(self):
        # Ignore this
        pass
    def getvalue(self):
        return self.sio.getvalue()

class wrap_stdin(object):
    def __init__(self, content):
        self.content = content
    def __enter__(self):
        self._old_stdin = sys.stdin
        sys.stdin = StringIO(self.content)
        return self
    def __exit__(self, *args):
        sys.stdin = self._old_stdin
    
# Given a filename in the "tests/" directory, return its full path

_dirname = os.path.dirname(__file__)
def fullpath(name):
    path = os.path.join(_dirname, name)
    assert os.path.exists(path), path
    return path

PUBCHEM_SDF = fullpath("pubchem.sdf")
PUBCHEM_SDF_GZ = fullpath("pubchem.sdf.gz")
PUBCHEM_ANOTHER_EXT = fullpath("pubchem.should_be_sdf_but_is_not")

MISSING_TITLE = fullpath("missing_title.sdf")

real_stdin = sys.stdin
real_stdout = sys.stdout
real_stderr = sys.stderr

class Runner(object):
    def __init__(self, main):
        self.main = main

    def pre_run(self):
        pass
    def post_run(self):
        pass

    def run(self, cmdline, source=PUBCHEM_SDF):
        if isinstance(cmdline, basestring):
            args = cmdline.split()
        else:
            args = cmdline
            assert isinstance(args, list) or isinstance(args, tuple)
        if source is not None:
            args = args + [source]
        self.pre_run()

        try:
            sys.stdout = stdout = SIO()
            self.main(args)
        finally:
            sys.stdout = real_stdout

        self.post_run()

        result = stdout.getvalue().splitlines()
        if result:
            self.verify_result(result)
        return result

    def verify_result(self, result):
        assert result[0] == "#FPS1", result[0]
        # TODO: .. verify more more line format ...

    def run_stdin(self, cmdline):
        raise NotImplementedError("Implement in the derived class")

    def run_fps(self, cmdline, expect_length=None, source=PUBCHEM_SDF):
        result = self.run(cmdline, source)
        while result[0].startswith("#"):
            del result[0]
        if expect_length is not None:
            assert len(result) == expect_length, (len(result), expect_length)
        return result

    def run_split(self, cmdline, expect_length=None, source=PUBCHEM_SDF):
        "split into dict of headers and list of values"
        result = self.run(cmdline, source)
        headers = {}
        fps = []
        result_iter = iter(result)
        # I know the first line is correct (it was tested in verify_result)
        # Plus, this lets the SimsearchRunner use run_split
        result_iter.next()
        for line in result_iter:
            if line.startswith("#"):
                k, v = line.split("=", 1)
                assert k not in headers, k
                headers[k] = v
                continue
            fps.append(line)
            break
        fps.extend(result_iter)
        if expect_length is not None:
            assert len(fps) == expect_length, (len(fps), expect_length)
        return headers, fps
            

    def run_exit(self, cmdline, source=PUBCHEM_SDF):
        sys.stderr = stderr = SIO()
        try:
            try:
                result = self.run(cmdline, source)
            except SystemExit:
                pass
            else:
                raise AssertionError("should have exited: %r gave: %r" % (cmdline, result))
        finally:
            sys.stderr = real_stderr
        return stderr.getvalue()

    def run_split_capture(self, cmdline, expect_length=None, source=PUBCHEM_SDF):
        sys.stderr = stderr = SIO()
        try:
            try:
                headers, fps = self.run_split(cmdline, expect_length, source)
            except SystemExit:
                raise AssertionError("unexpected SystemExit")
        finally:
            sys.stderr = real_stderr
        return headers, fps, stderr.getvalue()
        

####

def can_skip(name):
    s = os.environ.get("TOX_CHEMFP_TEST", "")
    return not (s.startswith(name) or (","+name) in s)

#### fingerprint encoding

def set_bit(n):
    assert n <= 16
    bytes = [0, 0, 0]
    bytes[n//8] = 1<<(n%8)
    return "%02x%02x%02x" % tuple(bytes)

class TestIdAndErrors(object):
    #
    # One of the records doesn't have an XLOGP field
    #
    def test_missing_id_tag(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag PUBCHEM_CACTVS_XLOGP", 18)
        self.assertNotIn("ERROR:", errmsg)
        self.assertNotIn("record #1", errmsg)
        self.assertNotIn("missing_title.sdf", errmsg)
        ids = [fp.split("\t")[1] for fp in fps]
        self.assertEquals(ids, ['2.8', '1.9', '1', '3.3', '1.5', '2.6', '-0.9', '2', '2.1', 
                                '2.9', '1.7', '-1.5', '0.4', '0.6', '0.4', '0.4', '2', '2.5'])

        
    def test_missing_id_strict(self):
        errmsg = self._runner.run_exit("--id-tag PUBCHEM_CACTVS_XLOGP --errors strict")
        self.assertIn("ERROR: Missing id tag 'PUBCHEM_CACTVS_XLOGP' in SD record", errmsg)
        self.assertIn("record #7", errmsg)
        self.assertIn("pubchem.sdf", errmsg)
    

    def test_missing_id_tag_report(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag PUBCHEM_CACTVS_XLOGP --errors report", 18)
        self.assertIn("ERROR: Missing title in SD record", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertIn("missing_title.sdf'", errmsg)
        self.assertEquals(fps[-1], "")

    # Should be the same as the default case.
    def test_missing_id_tag_ignore(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag PUBCHEM_CACTVS_XLOGP --errors ignore", 18)
        self.assertNotIn("ERROR:", errmsg)
        self.assertNotIn("record #1", errmsg)
        self.assertNotIn("missing_title.sdf", errmsg)
        ids = [fp.split("\t")[1] for fp in fps]
        self.assertEquals(ids, ['2.8', '1.9', '1', '3.3', '1.5', '2.6', '-0.9', '2', '2.1', 
                                '2.9', '1.7', '-1.5', '0.4', '0.6', '0.4', '0.4', '2', '2.5'])


    def test_missing_a_lot_of_ids_bypasses_ignore_errors(self):
        filename = get_tmpfile(self, "empty.smi")
        outfile = open(filename, "w")
        outfile.write("C\n" * 100)
        outfile.close()
        errmsg = self._runner.run_exit("--errors ignore", filename)
        self.assertIn(
            "Each of the first 100 records contained errors. Final error: Missing SMILES identifier (second column)",
            errmsg)
        self.assertIn("empty.smi'", errmsg)
        self.assertIn(", record #100", errmsg)
        self.assertIn("Exiting.", errmsg)
        
    #
    # Various ways of having a strange title
    #

    def test_missing_title(self):
        headers, fps, errmsg = self._runner.run_split_capture("", 1, MISSING_TITLE)
        self.assertIn("record #1", errmsg)
        self.assertIn("record #3", errmsg)
        self.assertEquals(len(fps), 1)
        self.assertEquals(fps[0].split(b"\t")[1], b"Good")

    def test_missing_title_strict(self):
        errmsg = self._runner.run_exit("--errors strict", MISSING_TITLE)
        self.assertIn("ERROR: Missing title in SD record", errmsg)
        self.assertIn("record #1", errmsg)

    def test_missing_title_report(self):
        headers, fps, errmsg = self._runner.run_split_capture("--errors report", 1, MISSING_TITLE)
        self.assertIn("ERROR: Missing title", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertNotIn("record #2", errmsg)
        self.assertIn("record #3", errmsg)
        self.assertEquals(len(fps), 1)
        self.assertEquals(fps[0].split(b"\t")[1], b"Good")

    def test_missing_title_ignore(self):
        headers, fps, errmsg = self._runner.run_split_capture("--errors ignore", 1, MISSING_TITLE)
        self.assertEqual(errmsg.count("record #1"), 1, errmsg)
        self.assertNotIn("record #2", errmsg)
        self.assertEqual(errmsg.count("record #3"), 1, errmsg)
        self.assertEquals(len(fps), 1)
        self.assertEquals(fps[0].split(b"\t")[1], b"Good")

    #
    # Various ways of handling a missing id in a tag
    #

    # Open Babel doesn't preserve the "    " so executes a different path
        
    def test_missing_id_tag(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Blank", 1, MISSING_TITLE)
        self.assertEqual(errmsg.count("ERROR"), 2, errmsg)
        self.assertEqual(errmsg.count("Empty id tag 'Blank' in SD record") +
                         errmsg.count("Missing id tag 'Blank' in SD record"), 2, errmsg)
        self.assertEqual(errmsg.count("record #"), 2, errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"This is not Blank")

    def test_missing_id_tag_strict(self):
        errmsg = self._runner.run_exit(
            "--id-tag Blank --errors strict", MISSING_TITLE)
        if self.toolkit == "openbabel":
            self.assertIn("ERROR: Missing id tag 'Blank' in SD record", errmsg)
        else:
            self.assertIn("ERROR: Empty id tag 'Blank' in SD record after cleanup", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertIn("missing_title.sdf'", errmsg)

    def test_missing_id_tag_report(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Blank --errors report", 1, MISSING_TITLE)
        if self.toolkit == "openbabel":
            self.assertIn("ERROR: Missing id tag 'Blank' in SD record", errmsg)
        else:
            self.assertIn("ERROR: Empty id tag 'Blank'", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertIn("record #2", errmsg)
        self.assertNotIn("record #3", errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"This is not Blank")

    def test_missing_id_tag_ignore(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Blank --errors ignore", 1, MISSING_TITLE)
        self.assertEqual(errmsg.count("ERROR"), 2, errmsg)
        self.assertEqual(errmsg.count("Missing id tag 'Blank'") +
                         errmsg.count("Empty id tag 'Blank'"), 2, errmsg)
        self.assertEqual(errmsg.count("record #"), 2, errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"This is not Blank")

    #
    # Various ways of handling a tab characters in an id tag
    #

    def test_tab_id_tag(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Tab", 2, MISSING_TITLE)
        ##  self.assertNotIn("ERROR: Empty id tag 'Tab'", errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"Leading tab")
        self.assertEquals(fps[1].split(b"\t")[1], b"This does not")

    def test_tab_id_tag_strict(self):
        errmsg = self._runner.run_exit(
            "--id-tag Tab --errors strict", MISSING_TITLE)
        if self.toolkit == "openbabel":
            self.assertIn("ERROR: Missing id tag 'Tab' in SD record", errmsg)
        else:
            self.assertIn("ERROR: Empty id tag 'Tab' in SD record after cleanup", errmsg)
        self.assertIn("record #2", errmsg)
        self.assertIn("missing_title.sdf'", errmsg)

    def test_tab_id_tag_report(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Tab --errors report", 2, MISSING_TITLE)
        if self.toolkit == "openbabel":
            self.assertIn("ERROR: Missing id tag 'Tab' in SD record", errmsg)
        else:
            self.assertIn("ERROR: Empty id tag 'Tab' in SD record after cleanup", errmsg)
        self.assertIn("record #2", errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"Leading tab")
        self.assertEquals(fps[1].split(b"\t")[1], b"This does not")

    def test_tab_id_tag_ignore(self):
        headers, fps, errmsg = self._runner.run_split_capture(
            "--id-tag Tab --errors ignore", 2, MISSING_TITLE)
        ##  self.assertNotIn("ERROR: Empty id tag 'Tab'", errmsg)
        self.assertEquals(fps[0].split(b"\t")[1], b"Leading tab")
        self.assertEquals(fps[1].split(b"\t")[1], b"This does not")


    def test_contains_tab_id_tag(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag ContainsTab", 3, MISSING_TITLE)
        self.assertNotIn("ERROR: Empty id tag 'ContainsTab'", errmsg)
        ids = [fp.split(b"\t")[1] for fp in fps]
        self.assertEquals(ids, [b"ThreeTabs", b"tabseparated", b"twotabs"])

    def test_contains_tab_id_tag_strict(self):
        headers, fps = self._runner.run_split("--id-tag ContainsTab --errors strict", 3, MISSING_TITLE)
        ids = [fp.split(b"\t")[1] for fp in fps]
        self.assertEquals(ids, [b"ThreeTabs", b"tabseparated", b"twotabs"])

    def test_contains_tab_id_tag_report(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag ContainsTab --errors report", 3, MISSING_TITLE)
        self.assertNotIn("ContainsTab", errmsg)
        self.assertNotIn("ERROR", errmsg)
        ids = [fp.split(b"\t")[1] for fp in fps]
        self.assertEquals(ids, [b"ThreeTabs", b"tabseparated", b"twotabs"])

    def test_contains_tab_id_tag_ignore(self):
        headers, fps, errmsg = self._runner.run_split_capture("--id-tag ContainsTab --errors ignore", 3, MISSING_TITLE)
        self.assertNotIn("ERROR: Empty id tag 'ContainsTab'", errmsg)
        ids = [fp.split(b"\t")[1] for fp in fps]
        self.assertEquals(ids, [b"ThreeTabs", b"tabseparated", b"twotabs"])

    #
    # Handling bad files
    #

    def test_handles_missing_filename(self):
        errmsg = self._runner.run_exit("this_file_does_not_exist.sdf", PUBCHEM_SDF)
        self.assertIn("Structure file '", errmsg)
        self.assertIn("this_file_does_not_exist.sdf", errmsg)
        self.assertIn("' does not exist", errmsg)
        self.assertNotIn("pubchem", errmsg)

    def test_handles_missing_filename_at_end(self):
        errmsg = self._runner.run_exit([PUBCHEM_SDF, "this_file_does_not_exist.sdf"])
        self.assertIn("Structure file '", errmsg)
        self.assertIn("this_file_does_not_exist.sdf", errmsg)
        self.assertIn("' does not exist", errmsg)
        self.assertNotIn("pubchem", errmsg)

    def test_unreadable_file(self):
        # Read three files. The second file is not readable.
        tf = tempfile.NamedTemporaryFile(suffix="unreadable.sdf")
        try:
            os.chmod(tf.name, 0o222)
            errmsg = self._runner.run_exit([PUBCHEM_SDF, tf.name])
            self.assertIn("Problem reading structure fingerprints", errmsg)
            self.assertIn("unreadable.sdf", errmsg)
            self.assertNotIn("pubchem", errmsg)
        finally:
            tf.close()

    def test_output_directory_does_not_exist(self):
        errmsg = self._runner.run_exit(["-o", "/never/never/land/exists/forever/xyzzy.fps", PUBCHEM_SDF])
        self.assertIn("Cannot open output fingerprint file:", errmsg)
        self.assertIn("/never/never/land/exists", errmsg)

    def test_bad_tag(self):
        result = self._runner.run_exit("--id-tag >nogood")
        self.assertIn("Invalid id tag: '>nogood'", result)

    def test_unknown_output_format(self):
        errmsg = self._runner.run_exit(["-o", "/another/path/that/does/not/exist.sdf.gz", PUBCHEM_SDF])
        self.assertIn("Cannot open output fingerprint file: Unable to determine fingerprint format type from",
                      errmsg)
        self.assertIn("does/not/exist", errmsg)
        self.assertIn("usage:", errmsg)
        
class TestIO(object):
    # '_runner' must be defined in the subclass
    def test_can_specify_input_format_matching_file(self):
        def without_source_header(cmdline, source):
            return [line for line in self._runner.run(cmdline, source)
                        if not line.startswith(b"#source=") and
                           not line.startswith(b"#date=")]
        result1 = without_source_header("", PUBCHEM_SDF)
        result2 = without_source_header("", PUBCHEM_SDF_GZ)
        self.assertEquals(result1, result2)

        result3 = without_source_header("--in sdf.gz", PUBCHEM_SDF_GZ)
        self.assertEquals(result1, result3)
        
        result4 = without_source_header("--in sdf", PUBCHEM_ANOTHER_EXT)
        self.assertEquals(result1, result4)
    
    
    def test_missing_filename(self):
        errmsg = self._runner.run_exit([], "does_not_exist.smi")
        self.assertIn("Structure file", errmsg)
        self.assertIn("does not exist", errmsg)
        self.assertIn("does_not_exist.smi", errmsg)

    def test_file_does_not_exist(self):
        msg = self._runner.run_exit([], source="/asdfaserwe.does.not.exist")
        self.assertIn("Structure file '/asdfaserwe.does.not.exist' does not exist", msg)

        
    def test_bad_input_extension(self):
        filename = get_tmpfile(self, "qwerty.xyzzy")
        open(filename, "wb").close()
        errmsg = self._runner.run_exit([], source=filename)
        self.assertIn("does not support the 'xyzzy' format", errmsg)
        
    def test_bad_input_flag(self):
        errmsg = self._runner.run_exit(" --in xyzzy")
        self.assertIn("Unsupported format specifier", errmsg)
        self.assertIn("xyzzy", errmsg)

    def test_unreadable_input(self):
        filename = get_tmpfile(self, "unreadable.smi")
        open(filename, "wb").close()
        os.chmod(filename, 0)
        try:
            open(filename)
        except IOError as err:
            expected_msg = str(err)
            
        errmsg = self._runner.run_exit([], source=filename)
        self.assertIn("Problem reading structure fingerprints", errmsg)
        self.assertIn(expected_msg, errmsg)

    def test_bad_output_extension(self):
        filename = get_tmpfile(self, "bad_extension.fpq")
        errmsg = self._runner.run_exit(["--output", filename])
        self.assertIn("Cannot open output fingerprint file: Unable to determine fingerprint format type from", errmsg)
        self.assertIn("bad_extension.fpq", errmsg)
        self.assertIn("\nUse --out to specify 'fps' or 'fps.gz'.", errmsg)
        
    def test_bad_output_flag(self):
        errmsg = self._runner.run_exit(["--out", "fpq"])
        self.assertIn("invalid choice: 'fpq' (choose from 'fps', 'fps.gz', 'flush')", errmsg)

    def test_fps_output(self):
        filename = get_tmpfile(self, "output.xyz")
        self._runner.run(["--out", "fps", "-o", filename])
        with open(filename) as infile:
            self.assertEqual(next(infile), "#FPS1\n")
            
    def test_fps_gz_output(self):
        filename = get_tmpfile(self, "output.xyz")
        self._runner.run(["--out", "fps.gz", "-o", filename])
        import gzip
        with gzip.open(filename) as infile:
            self.assertEqual(next(infile), b"#FPS1\n")

    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_output(self):
        filename = get_tmpfile(self, "output.xyz")
        self._runner.run(["--out", "flush", "-o", filename])
        from chemfp_converters import flush
        with flush.open_flush(filename) as reader:
            self.assertIsNot(reader.metadata, None)
        
    @unittest2.skipIf(has_chemfp_converters, "chemfp_converters is installed")
    def test_flush_output_no_converter(self):
        filename = get_tmpfile(self, "output.xyz")
        output = self._runner.run_exit(["--out", "flush", "-o", filename])
        self.assertEqual(output, "--out format 'flush' not supported because the chemfp_converter module is not available\n")
        
    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_filename(self):
        filename = get_tmpfile(self, "output.flush")
        self._runner.run(["-o", filename])
        from chemfp_converters import flush
        with flush.open_flush(filename) as reader:
            self.assertIsNot(reader.metadata, None)
        
    @unittest2.skipIf(has_chemfp_converters, "chemfp_converters is installed")
    def test_flush_filename_no_converter(self):
        filename = get_tmpfile(self, "output.flush")
        output = self._runner.run_exit(["-o", filename])
        self.assertIn("Cannot open output fingerprint file: "
                      "Cannot write to flush files because the chemfp_converter module is not available\n",
                      output)
        
        

def get_tmpdir(test_case):
    dirname = tempfile.mkdtemp()
    test_case.addCleanup(shutil.rmtree, dirname)
    return dirname

def get_tmpfile(testcase, filename):
    dirname = tempfile.mkdtemp(prefix="chemfp_test")
    testcase.addCleanup(shutil.rmtree, dirname)
    filename = os.path.join(dirname, filename)
    return filename



import unittest2
import sys
import os
from cStringIO import StringIO as SIO

from chemfp.bitops import hex_contains_bit
import support

try:
    from openeye import oechem  # These tests require OEChem
    if not oechem.OEChemIsLicensed():
        print >>sys.stderr, "oechem library available but not license found. Skipping its tests."
        raise ImportError
    has_oechem = True
    skip_oechem = False
except ImportError:
    has_oechem = False
    skip_oechem = True
    if not support.can_skip("oe"):
        skip_oechem = False
        from openeye import oechem

if has_oechem:
    from chemfp.commandline import oe2fps
    import chemfp.openeye
    chemfp.openeye._USE_SELECT = False # Grrr. Needed to automate testing.

    real_stdout = sys.stdout
    real_stderr = sys.stderr

    PUBCHEM_SDF = support.fullpath("pubchem.sdf")
    PUBCHEM_SDF_GZ = support.fullpath("pubchem.sdf.gz")
    PUBCHEM_ANOTHER_EXT = support.fullpath("pubchem.should_be_sdf_but_is_not")

    MACCS_SMI = support.fullpath("maccs.smi")

    oeerrs = oechem.oeosstream()
    oechem.OEThrow.SetOutputStream(oeerrs)

def _check_for_oe_errors():
    lines = oeerrs.str().splitlines()
    for line in lines:
        if line.startswith("Warning: Stereochemistry corrected on atom number"):
            continue
        if line.startswith("Warning: Unknown file format set in input stream"):
            # There's a bug in OEChem where it generates this warning on unknown
            # file extensions even after SetFormat has been called
            continue
        raise AssertionError("Unexpected message from OEChem: %r" % (line,))


# I build the fingerprints using bit offsets to ensure that the test
# data matches the actual bit results from OEChem. While I could
# reproduce the method in chemfp.openeye.get_maccs_fingerprinter, that
# would be cheating. I could test against ToHexString() but then I
# would have the nagging feeling that I got the ordering
# backwards. Instead, I do it from scratch using the bit offset.

def _construct_test_values(fp_func = None, num_bits=4096):
    from openeye.oechem import oemolistream
    from openeye.oegraphsim import OEFingerPrint, OEMakePathFP
    fp = OEFingerPrint()
    ifs = oemolistream()
    assert ifs.open(PUBCHEM_SDF)
    hex_data = []
    if fp_func is None:
        fp_func = OEMakePathFP

    def _convert_to_chemfp_order(s):
        # The FPS format allows either case but I prefer lowercase
        s = s.lower()
        # OpenEye orders hex values on nibbles. Chemfp orders on bytes.
        return "".join( (s[i+1]+s[i]) for i in range(0, len(s), 2))

    for mol in ifs.GetOEGraphMols():
        fp_func(fp, mol)
        # Set the byte values given the bit offsets
        bytes = [0] * (num_bits//8)
        i = fp.FirstBit()
        while i >= 0:
            bytes[i//8] |= 1<<(i%8)
            i = fp.NextBit(i)
        as_hex = "".join("%02x" % i for i in bytes)
        assert len(as_hex) == 2*(num_bits//8), len(as_hex)
        # Double-check that it matches the (reordered) ToHexString()
        oe_hex = fp.ToHexString()[:-1]
        assert as_hex == _convert_to_chemfp_order(oe_hex), (
            as_hex, _convert_to_chemfp_order(oe_hex))
        
        hex_data.append("%s\t%s" % (as_hex, mol.GetTitle()))
    return hex_data

# I have this to flag any obvious changes in the OEChem algorithm and
# to help with figuring out how to build a test case.

_fp1 = "00001002200200000000000000000000000008400020000300801002300000000200000840000000000080000000000000204008000000000010000c10000000400000010100000210800002000000009400000000020020088000000000010000918000200000580400002000010020002440000008001001404000000200010000a8c00020400200002000004084000000030100820000000000000002000000510001800000010001000081100110000800480000100400000c00004c000800000808000100000022000228800020004000000200182100000100000000101000010004004808000000800000000001010010201000000090400000100000020010000010201000000040300100000000580000000000000200000000401000000000000008040004000000008002080820000310280200004040a000000010000080005000004010010018000000800000020008208040000400000200000000000000000800080050000008400100004000000200ac001000000000800100200060900010002000000040200000000000040808000048400040000000020000001001000000000302002008200000a044000180800000100000000200000049004080080000100022a00084000400280480000000402400080400404100000000040000020c10000000000c000100002000080010002080100002000600"

if has_oechem:
    hex_test_values = _construct_test_values()
    assert hex_test_values[0].startswith(_fp1)

class OERunner(support.Runner):
    def pre_run(self):
        oeerrs.clear()
    def post_run(self):
        _check_for_oe_errors()

    def run_stdin(self, cmdline, source):
        fd = os.open(source, os.O_RDONLY)
        oechem.oein.openfd(fd, False)
        try:
            return self.run(cmdline, None)
        finally:
            oechem.oein.openfd(0, False)
            os.close(fd)

if has_oechem:
    runner = OERunner(oe2fps.main)
    run = runner.run
    run_stdin = runner.run_stdin
    run_fps = runner.run_fps
    run_exit = runner.run_exit
else:
    runner = None

def headers(lines):
    assert lines[0] == "FPS1"
    del lines[0]
    return [line for line in lines if line.startswith("#")]


class TestMACCS(unittest2.TestCase):
    def test_bitorder(self):
        result = run_fps("--maccs166", 7, support.fullpath("maccs.smi"))
        # The fingerprints are constructed to test the first few bytes.
        self.assertEquals(result[0][:6], support.set_bit(2))
        self.assertEquals(result[1][:6], support.set_bit(3))
        self.assertEquals(result[2][:6], support.set_bit(4))
        self.assertEquals(result[3][:6], support.set_bit(5))
        self.assertEquals(result[4][:6], support.set_bit(9))
        self.assertEquals(result[5][:6], support.set_bit(10))
        self.assertEquals(result[6][:6], support.set_bit(16))

TestMACCS = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestMACCS)

class TestPath(unittest2.TestCase):
    def test_default(self):
        result = run_fps("", 19)
        hexfp, id = result[0].split()
        self.assertEquals(len(hexfp), 4096//4)
        self.assertEquals(result[0], hex_test_values[0])
        self.assertEquals(result, hex_test_values)

    def test_path_option(self):
        result = run_fps("--path", 19)
        self.assertEquals(result, hex_test_values)

    @unittest2.skipIf(has_oechem and oechem.OEChemGetRelease() == "2.1.3.b.1 debug",
                      "skip an assertion failure in this debug version")
    def test_num_bits(self):
        result = run_fps("--numbits 16", 19)
        self.assertEquals(result[0][:5], "ff1f\t")

    def test_min_bonds_default(self):
        result = run_fps("--minbonds 0", 19)
        self.assertEquals(result, hex_test_values)

    def test_min_bonds_changed(self):
        result = run_fps("--minbonds 1", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_max_bonds_default(self):
        result = run_fps("--maxbonds 5", 19)
        self.assertEquals(result, hex_test_values)
        
    def test_max_bonds_changed(self):
        result = run_fps("--minbonds 4", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_default_named(self):
        result = run_fps("--atype DefaultAtom", 19)
        self.assertEquals(result, hex_test_values)

    def test_atype_default_flags(self):
        result = run_fps(
            "--atype Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb", 19)
        self.assertEquals(result, hex_test_values)


    def test_atype_default_flags_with_duplicates(self):
        result = run_fps(
            "--atype Arom|Chiral|AtmNum|EqHalo|HvyDeg|FCharge|Hyb", 19)
        self.assertEquals(result, hex_test_values)

    # Make sure that each of the flags returns some other answer
    def test_atype_different_1(self):
        result = run_fps(
            "--atype AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_2(self):
        result = run_fps(
            "--atype Arom|Chiral|EqHalo|FCharge|HvyDeg|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_3(self):
        result = run_fps(
            "--atype Arom|AtmNum|EqHalo|FCharge|HvyDeg|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_4(self):
        result = run_fps(
            "--atype Arom|AtmNum|Chiral|FCharge|HvyDeg|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_5(self):
        result = run_fps(
            "--atype Arom|AtmNum|Chiral|EqHalo|HvyDeg|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_6(self):
        result = run_fps(
            "--atype Arom|AtmNum|Chiral|EqHalo|FCharge|Hyb", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_atype_different_7(self):
        result = run_fps(
            "--atype Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg", 19)
        self.assertNotEquals(result, hex_test_values)


    def test_btype_default_named(self):
        result = run_fps("--btype DefaultBond", 19)
        self.assertEquals(result, hex_test_values)

    def test_btype_default_flags(self):
        result = run_fps("--btype Order|Chiral", 19)
        self.assertEqual(result, hex_test_values)

    def test_btype_different_1(self):
        result = run_fps("--btype Order", 19)
        self.assertNotEquals(result, hex_test_values)

    def test_btype_different_2(self):
        result = run_fps("--btype Chiral", 19)
        self.assertNotEquals(result, hex_test_values)

TestPath = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestPath)

class TestPatterns(unittest2.TestCase):
    def test_rdmaccs(self):
        headers, fps = runner.run_split("--rdmaccs", 19)
        self.assertEqual(headers[b"#type"], b"RDMACCS-OpenEye/2")
        self.assertEqual(fps[0], b"000000000002080019c444eacd6c981baea178ef1f\t9425004")
        self.assertEqual(fps[1], b"000000002000082159d404eea968b81b8ea17eef1f\t9425009")
        self.assertEqual(fps[2], b"000000000000080159c404efa9689a1b8eb1faef1b\t9425012")
        self.assertEqual(fps[3], b"000000000000082019c404ee8968b81b8ea1ffef1f\t9425015")
        self.assertEqual(fps[4], b"000000000000088419c6b5fa8968981b8eb37aef1f\t9425018")

    def test_rdmaccs1_key44(self):
        header, output = runner.run_split("--rdmaccs/1", 7, source=MACCS_SMI)
        self.assertEqual(header[b"#type"], b"RDMACCS-OpenEye/1")
        hex_fp, id = output[0].split(b"\t")
        self.assertFalse(hex_contains_bit(hex_fp, 43))
        self.assertEqual(output[0], b"040000000000000000000000000000000000000000\t3->bit_2")
        
    def test_rdmaccs2_key44(self):
        header, output = runner.run_split("--rdmaccs/2", 7, source=MACCS_SMI)
        self.assertEqual(header[b"#type"], b"RDMACCS-OpenEye/2")
        hex_fp, id = output[0].split(b"\t")
        self.assertTrue(hex_contains_bit(hex_fp, 43))
        self.assertEqual(output[0], b"040000000008000000000000000000000000000000\t3->bit_2")
        #                   There's an 8 here ---^
    
    def test_both_rdmaccs_flags_gives_error(self):
        result = runner.run_exit("--rdmaccs/1 --rdmaccs/2")
        self.assertIn("Cannot specify both --rdmaccs and --rdmaccs/1", result)
        result = runner.run_exit("--rdmaccs --rdmaccs/1")
        self.assertIn("Cannot specify both --rdmaccs and --rdmaccs/1", result)
        runner.run_split("--rdmaccs --rdmaccs/2", 7, source=MACCS_SMI)
        
    def test_substruct(self):
        headers, fps = runner.run_split("--substruct", 19)
        self.assertEquals(headers["#type"], "ChemFP-Substruct-OpenEye/1")
        self.assertEquals(fps[0], "07de8d002000000000000000000000000080060000000c000000000000000080030000f8401800000030508379344c014956000055c0a44e2a0049200084e140581f041d661b10064483cb0f2925100619001393e10001007000000000008000000000000000400000000000000000\t9425004")
        # Note: not the same as OpenBabel's answer; bit 260 (>= 3 hetero-aromatic rings) is different.
        # openeye_patterns doesn't handle this.
        self.assertEquals(fps[1], "07de0d000000000000000000000000000080460300000c000000000000000080070000780038000000301083f920cc09695e0800d5c0e44e6e00492190844145dc1f841d261911164d039b8f29251026b9401313e0ec01007000000000000000000000000000000000000000000000\t9425009")

TestPatterns = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestPatterns)

@unittest2.skipIf(skip_oechem, "OEChem not installed")
class TestIdAndErrors(unittest2.TestCase, support.TestIdAndErrors):
    _runner = runner
    toolkit = "openeye"

@unittest2.skipIf(skip_oechem, "OEChem not installed")
class TestIO(unittest2.TestCase, support.TestIO):
    _runner = runner
    def test_compressed_input(self):
        result = run_fps("", source=PUBCHEM_SDF_GZ)
### XXX Fix how I handle unknown extensions. 
#    def test_unknown_extension(self):
#        # OEChem's default assumes SMILES. This will parse some of the
#        # SD file lines as SMILES and skip the ones it doesn't know.
#        # The error output will have a bunch of warnings, starting
#        # with the "Unknown file format ... " warning, and then this
#        # string about a SMILES parse error.
#        try:
#            run("--errors ignore", source=PUBCHEM_ANOTHER_EXT)
#        except AssertionError, x:
#            self.assertEquals("Problem parsing SMILES" in str(x), True, str(x))
            
    def test_specify_input_format(self):
        result = run_fps("--in sdf", source=PUBCHEM_ANOTHER_EXT)


    def test_from_stdin(self):
        run_stdin("--in sdf", source=PUBCHEM_SDF)

    def test_from_gziped_stdin(self):
        run_stdin("--in sdf.gz", source=PUBCHEM_SDF_GZ)


# XXX how to test that this generates a warning?
#    def test_specify_input_format_with_dot(self):
#        result = run_fps("--in .sdf", source=PUBCHEM_ANOTHER_EXT)

class TestArgErrors(unittest2.TestCase):
    def _run(self, cmd, expect):
        msg = run_exit(cmd)
        self.assertIn(expect, msg)

    def test_two_fp_types(self):
        self._run("--maccs166 --path", "Cannot specify both --maccs166 and --path")

    def test_num_bits_too_small(self):
        self._run("--numbits 0", "between 16 and 65536 bits")
        self._run("--numbits 1", "between 16 and 65536 bits")
        self._run("--numbits 15", "between 16 and 65536 bits")

    def test_num_bits_too_large(self):
        self._run("--numbits 65537", "between 16 and 65536 bits")
        # Check for overflow, even though I know it won't happen in Python
        self._run("--numbits %(big)s"%dict(big=2**32+32), "between 16 and 65536 bits")

    def test_min_bonds_too_small(self):
        self._run("--minbonds=-1", "0 or greater")

    def test_min_bonds_larger_than_default_max_bonds(self):
        self._run("--minbonds=6", "--maxbonds must not be smaller than --minbonds")

    def test_min_bonds_too_large(self):
        self._run("--minbonds=4 --maxbonds=3",
                  "--maxbonds must not be smaller than --minbonds")

    def test_bad_atype(self):
        self._run("--atype spam", "Unknown path atom type 'spam'")

    def test_bad_atype2(self):
        self._run("--atype DefaultAtom|spam", "Unknown path atom type 'spam'")

    def test_bad_atype3(self):
        self._run("--atype DefaultAtom|", "Missing path atom flag")

    def test_bad_btype(self):
        self._run("--btype eggs", "Unknown path bond type 'eggs'")

    def test_bad_btype2(self):
        self._run("--btype DefaultBond|eggs", "Unknown path bond type 'eggs'")

    def test_bad_btype3(self):
        self._run("--btype DefaultBond|", "Missing path bond flag")

TestArgErrors = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestArgErrors)

class TestHeaderOutput(unittest2.TestCase):
    def _field(self, s, field):
        try:
            result = run(s)
        except SystemExit, err:
            raise
            raise AssertionError("Should not die: %r" % (err,))
        filtered = [line for line in result if line.startswith(field)]
        self.assertEquals(len(filtered), 1, result)
        return filtered[0]

    def test_software(self):
        result = self._field("", "#software")
        self.assertEquals("#software=OEGraphSim/" in result, True, result)
        self.assertIn("(", result)
        self.assertIn(")", result)
        result = self._field("--maccs166", "#software")
        self.assertIn("#software=OEGraphSim/", result)

    def test_type(self):
        result = self._field("", "#type")
        self.assertEqual(result, 
                b"#type=OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 "
                b"atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")

    def test_default_atom_and_bond(self):
        result = self._field(
            "--atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb "
            "--btype=Order|Chiral", b"#type")
        self.assertEqual(result,
            b"#type=OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 "
            b"atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")

        
    # different flags. All flags? and order
    def test_num_bits(self):
        result = self._field("--numbits 38", b"#num_bits")
        self.assertEqual(result, b"#num_bits=38")
        
    def test_atype_flags(self):
        result = self._field("--atype FCharge|FCharge", b"#type") + b" "
        self.assertIn(b" atype=FCharge ", result)
    
    def test_btype_flags(self):
        result = self._field("--btype Chiral|Order", b"#type") + b" "
        self.assertIn(b" btype=Order|Chiral ", result)
        result = self._field("--btype Order|Chiral", b"#type") + b" "
        self.assertIn(b" btype=Order|Chiral ", result)
    
    def test_pipe_or_comma(self):
        result = self._field("--atype HvyDeg,FCharge --btype Chiral,Order",
                             b"#type") + b" "
        self.assertIn(b" atype=FCharge|HvyDeg ", result)
        self.assertIn(b" btype=Order|Chiral ", result)
        
    def test_maccs_header(self):
        name = "OpenEye-MACCS166/" + chemfp.openeye._maccs_version
        result = self._field("--maccs166", b"#type")
        self.assertEqual(result, b"#type=" + name.encode("ascii"))
        
TestHeaderOutput = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestHeaderOutput)


if has_oechem:
    from openeye.oegraphsim import (
        OEMakePathFP, OEFPAtomType_DefaultPathAtom, OEFPBondType_DefaultPathBond,
        OEMakeCircularFP, OEFPAtomType_DefaultCircularAtom, OEFPBondType_DefaultCircularBond,
        OEMakeTreeFP, OEFPAtomType_DefaultTreeAtom, OEFPBondType_DefaultTreeBond,
        OEFPAtomType_Aromaticity, OEFPAtomType_AtomicNumber, OEFPAtomType_EqHalogen,
        OEFPAtomType_HvyDegree, OEFPAtomType_FormalCharge,
        OEFPBondType_Chiral, OEFPBondType_InRing, OEFPBondType_BondOrder,
        )


class TestOEGraphSimVersion2(unittest2.TestCase):
    def test_hash(self):
        header, result = runner.run_split("--path", 19)
        self.assertEquals(header["#type"], "OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")
        self.assertEquals(result, _construct_test_values())
        
    def test_path_defaults(self):
        header, result = runner.run_split("--path --numbits 4096 --minbonds 0 --maxbonds 5 "
                                          "--atype AtmNum|Arom|Chiral|FCharge|HvyDeg|Hyb|EqHalo --btype Order|Chiral", 19)
        self.assertEquals(header["#type"], "OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")
        self.assertEquals(result, _construct_test_values())

    def test_change_all_path_fields(self):
        header, result = runner.run_split("--path --numbits 1024 --minbonds 2 --maxbonds 4 "
                                          "--atype AtmNum|EqHalo --btype InRing|Order", 19)
        self.assertEquals(header["#type"], "OpenEye-Path/2 numbits=1024 minbonds=2 maxbonds=4 atype=AtmNum|EqHalo btype=Order|InRing")
        def compute_path_fingerprints(fp, mol):
            OEMakePathFP(fp, mol, 1024, 2, 4,
                         OEFPAtomType_AtomicNumber|OEFPAtomType_EqHalogen,
                         OEFPBondType_InRing|OEFPBondType_BondOrder)
        self.assertEquals(result, _construct_test_values(compute_path_fingerprints, 1024))
        
    def test_path_default_type(self):
        result = run("--atype DefaultPathAtom") # (DefaultPathAtom is the same as DefaultAtom)
        typename = [line for line in result if line.startswith("#type=")][0]
        self.assertEquals(typename, "#type=OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")

    def test_check_path_default_types(self):
        header, result = runner.run_split("--path --atype Default --btype Default")
        self.assertEquals(header["#type"],
              "OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")
        header, result = runner.run_split("--path --atype DefaultCircularAtom --btype DefaultCircularBond")
        self.assertEquals(header["#type"],
              "OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HCount btype=Order")
        header, result = runner.run_split("--path --atype DefaultPathAtom --btype DefaultPathBond")
        self.assertEquals(header["#type"],
              "OpenEye-Path/2 numbits=4096 minbonds=0 maxbonds=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")


    ########################
    # Note: The documentation says that OEFPBondType_DefaultCircularBond is Order|Chiral
    # but the code says it's only Order.
        
    def test_circular(self):
        header, result = runner.run_split("--circular", 19)
        def compute_circular_fingerprints(fp, mol):
            OEMakeCircularFP(fp, mol, 4096, 0, 5,
                             OEFPAtomType_DefaultCircularAtom, OEFPBondType_DefaultCircularBond)
        self.assertEquals(header["#type"], "OpenEye-Circular/2 numbits=4096 minradius=0 maxradius=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HCount btype=Order")
        self.assertEquals(result, _construct_test_values(compute_circular_fingerprints))

    def test_circular_defaults(self):
        # Make sure that when I specify the defaults then I get the same results
        header, result = runner.run_split("--circular --numbits 4096 --minradius 0 --maxradius 5 "
                                          "--atype AtmNum|Arom|Chiral|FCharge|HCount|EqHalo --btype Order")
        self.assertEquals(header["#type"], "OpenEye-Circular/2 numbits=4096 minradius=0 maxradius=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HCount btype=Order")
                
        def compute_circular_fingerprints(fp, mol):
            OEMakeCircularFP(fp, mol, 4096, 0, 5,
                             OEFPAtomType_DefaultCircularAtom, OEFPBondType_DefaultCircularBond)
        self.assertEquals(result, _construct_test_values(compute_circular_fingerprints))

    def test_change_all_circular_fields(self):
        header, result = runner.run_split("--circular --numbits 1024 --minradius 2 --maxradius 4 --atype Arom --btype Chiral", 19)
        def compute_circular_fingerprints(fp, mol):
            OEMakeCircularFP(fp, mol, 1024, 2, 4,
                             OEFPAtomType_Aromaticity, OEFPBondType_Chiral)
        self.assertEquals(header["#type"],
              "OpenEye-Circular/2 numbits=1024 minradius=2 maxradius=4 atype=Arom btype=Chiral")
        self.assertEquals(result, _construct_test_values(compute_circular_fingerprints, 1024))

    def test_check_circular_default_types(self):
        header, result = runner.run_split("--circular --atype Default --btype Default")
        self.assertEquals(header["#type"],
              "OpenEye-Circular/2 numbits=4096 minradius=0 maxradius=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HCount btype=Order")
        header, result = runner.run_split("--circular --atype DefaultCircularAtom --btype DefaultCircularBond")
        self.assertEquals(header["#type"],
              "OpenEye-Circular/2 numbits=4096 minradius=0 maxradius=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HCount btype=Order")
        header, result = runner.run_split("--circular --atype DefaultPathAtom --btype DefaultPathBond")
        self.assertEquals(header["#type"],
              "OpenEye-Circular/2 numbits=4096 minradius=0 maxradius=5 atype=Arom|AtmNum|Chiral|EqHalo|FCharge|HvyDeg|Hyb btype=Order|Chiral")


    ########################

    def test_tree(self):
        header, result = runner.run_split("--tree", 19)
        self.assertEquals(header["#type"], "OpenEye-Tree/2 numbits=4096 minbonds=0 maxbonds=4 atype=Arom|AtmNum|Chiral|FCharge|HvyDeg|Hyb btype=Order")
        def compute_tree_fingerprints(fp, mol):
            OEMakeTreeFP(fp, mol, 4096, 0, 4,
                         OEFPAtomType_DefaultTreeAtom, OEFPBondType_DefaultTreeBond)
        self.assertEquals(result, _construct_test_values(compute_tree_fingerprints))
    
    def test_tree_defaults(self):
        # Make sure that when I specify the defaults then I get the same results
        header, result = runner.run_split("--tree --numbits 4096 --minradius 0 --maxradius 5 "
                                          "--atype FCharge|HvyDeg|AtmNum|Arom|Chiral|Hyb --btype Order")
        self.assertEquals(header["#type"], "OpenEye-Tree/2 numbits=4096 minbonds=0 maxbonds=4 atype=Arom|AtmNum|Chiral|FCharge|HvyDeg|Hyb btype=Order")
        def compute_tree_fingerprints(fp, mol):
            OEMakeTreeFP(fp, mol, 4096, 0, 4,
                         OEFPAtomType_DefaultTreeAtom, OEFPBondType_DefaultTreeBond)
        self.assertEquals(result, _construct_test_values(compute_tree_fingerprints))

    
    def test_change_all_tree_fields(self):
        header, result = runner.run_split("--tree --numbits 1024 --minbonds 1 --maxbonds 2 --atype HvyDeg|FCharge --btype InRing", 19)
        def compute_circular_fingerprints(fp, mol):
            OEMakeTreeFP(fp, mol, 1024, 1, 2,
                         OEFPAtomType_HvyDegree | OEFPAtomType_FormalCharge, OEFPBondType_InRing)
        self.assertEquals(header["#type"],
              "OpenEye-Tree/2 numbits=1024 minbonds=1 maxbonds=2 atype=FCharge|HvyDeg btype=InRing")
        self.assertEquals(result, _construct_test_values(compute_circular_fingerprints, 1024))
    
if skip_oechem:
    TestOEGraphSimVersion2 = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestOEGraphSimVersion2)
        
if __name__ == "__main__":
    unittest2.main()



import sys
import unittest2

from chemfp.commandline import sdf2fps
from chemfp.bitops import hex_encode_as_bytes
import io

import support

real_stdin = sys.stdin
real_stdout = sys.stdout
real_stderr = sys.stderr

DECODER_SDF = support.fullpath("decoder.sdf")
DECODER_SDF_BYTES = open(DECODER_SDF, "rb").read()

PUBCHEM_SDF = support.fullpath("pubchem.sdf")
PUBCHEM_SDF_BYTES = open(PUBCHEM_SDF, "rb").read()
PUBCHEM_SDF_GZ = support.fullpath("pubchem.sdf.gz")

_runner = support.Runner(sdf2fps.main)
run_exit = _runner.run_exit

run = _runner.run
run_split_capture = _runner.run_split_capture


def run_fps(s, expect_length=None):
    result = run(s, DECODER_SDF)
    while result[0].startswith(b"#"):
        del result[0]
    if expect_length is not None:
        assert len(result) == expect_length
    return result


class TestDecoderFlags(unittest2.TestCase):
    def test_cactvs(self):
        result = run_fps("--cactvs --fp-tag PUBCHEM_CACTVS_SUBSKEYS")
        self.assertEquals(result, [b"07de8d002000000000000000000000000080060000000c000000000000000080030000f8401800000030508379344c014956000055c0a44e2a0049200084e140581f041d661b10064483cb0f2925100619001393e10001007000000000008000000000000000400000000000000000\t9425004",
                                   b"07de0d000000000000000000000000000080460300000c0000000000000000800f0000780038000000301083f920cc09695e0800d5c0e44e6e00492190844145dc1f841d261911164d039b8f29251026b9401313e0ec01007000000000000000000000000000000000000000000000\t9425009"])

    def test_binary40(self):
        result = run_fps("--binary --fp-tag binary40", 2)
        self.assertEquals(result[0], b"000500c000\t9425004")
        self.assertEquals(result[1], b"00fab75300\t9425009")

    def test_binary_msb40(self):
        result = run_fps("--binary-msb --fp-tag binary40", 2)
        self.assertEquals(result[0], b"000300a000\t9425004")
        self.assertEquals(result[1], b"00caed5f00\t9425009")

    def test_binary3(self):
        result = run_fps("--binary --fp-tag binary3", 2)
        self.assertEquals(result[0], b"04\t9425004")
        self.assertEquals(result[1], b"03\t9425009")

    def test_binary_msb3(self):
        result = run_fps("--binary-msb --fp-tag binary3", 2)
        self.assertEquals(result[0], b"01\t9425004")
        self.assertEquals(result[1], b"06\t9425009")

    def test_binary8(self):
        result = run_fps("--binary --fp-tag binary8", 2)
        self.assertEquals(result[0], b"76\t9425004")
        self.assertEquals(result[1], b"bc\t9425009")

    def test_binary_msb8(self):
        result = run_fps("--binary-msb --fp-tag binary8", 2)
        self.assertEquals(result[0], b"6e\t9425004")
        self.assertEquals(result[1], b"3d\t9425009")


    def test_binary17(self):
        result = run_fps("--binary --fp-tag binary17", 2)
        self.assertEquals(result[0], b"38b701\t9425004")
        self.assertEquals(result[1], b"489d01\t9425009")

    def test_binary_msb17(self):
        result = run_fps("--binary-msb --fp-tag binary17", 2)
        self.assertEquals(result[0], b"db3900\t9425004")
        self.assertEquals(result[1], b"732500\t9425009")


    def test_binary_failure(self):
        errmsg = run_exit("--binary --fp-tag PUBCHEM_CACTVS_SUBSKEYS", DECODER_SDF)
        self.assertIn("Could not binary decode tag 'PUBCHEM_CACTVS_SUBSKEYS' value 'AAADceB7sQ", errmsg)
        self.assertIn("Must pass in a string containing only 0s and 1s", errmsg)
        self.assertIn("line 1", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertIn("decoder.sdf", errmsg)

    def test_binary_msb_failure(self):
        errmsg = run_exit("--binary-msb --fp-tag PUBCHEM_CACTVS_SUBSKEYS", DECODER_SDF)
        self.assertIn("Could not binary_msb decode tag 'PUBCHEM_CACTVS_SUBSKEYS' value 'AAADceB7sQ", errmsg)
        self.assertIn("Must pass in a string containing only 0s and 1s", errmsg)
        self.assertIn("line 1", errmsg)
        self.assertIn("record #1", errmsg)
        self.assertIn("decoder.sdf", errmsg)


    def test_hex2(self):
        result = run_fps("--hex --fp-tag hex2", 2)
        self.assertEquals(result[0], b"ab\t9425004")
        self.assertEquals(result[1], b"01\t9425009")
        
    def test_hex_lsb2(self):
        # 0xab == 0b10101011
        # 10101011 with LSB first is 5 d => "d5"
        # 0x01 == 0b00000001 => 80 when in LSB first
        result = run_fps("--hex-lsb --fp-tag hex2", 2)
        self.assertEquals(result[0], b"d5\t9425004")
        self.assertEquals(result[1], b"80\t9425009")

    def test_hex_msb2(self):
        # With 2 nibbles the result is the same as hex
        result = run_fps("--hex-msb --fp-tag hex2", 2)
        self.assertEquals(result[0], b"ab\t9425004")
        self.assertEquals(result[1], b"01\t9425009")

    def test_hex16(self):
        result = run_fps("--hex --fp-tag hex16", 2)
        self.assertEquals(result[0], b"0123456789abcdef\t9425004")
        self.assertEquals(result[1], b"abcdef0123456789\t9425009")
        
    def test_hex_lsb16(self):
        result = run_fps("--hex-lsb --fp-tag hex16", 2)
        # 0123456789abcdef in LSB form => 
        # 084c2a6e195d3b7f when nibbles bits are in MSB form but nibbles are LSB
        # 80 c4 a2 e6 91 d5 b3 f7 when byte bits are in MSB and bytes are LSB
        self.assertEquals(result[0], b"80c4a2e691d5b3f7\t9425004")
        # abcdef0123456789 in LSB form =>
        # 5d3b7f084c2a6e19 =>
        # d5 b3 f7 80 c4 a2 e6 91
        self.assertEquals(result[1], b"d5b3f780c4a2e691\t9425009")

    def test_hex_msb16(self):
        # Just a bit of reordering
        result = run_fps("--hex-msb --fp-tag hex16", 2)
        self.assertEquals(result[0], b"efcdab8967452301\t9425004")
        self.assertEquals(result[1], b"8967452301efcdab\t9425009")
        
    def test_base64_16(self):
        result = run_fps("--base64 --fp-tag base64_16", 2)
        self.assertEquals(result[0], hex_encode_as_bytes("Greetings, human") + b"\t9425004")
        self.assertEquals(result[1], hex_encode_as_bytes("blahblahspamblah") + b"\t9425009")

    def test_daylight1(self):
        result = run_fps("--daylight --fp-tag daylight1", 2)
        self.assertEquals(result[0], hex_encode_as_bytes("PyDaylight") + b"\t9425004")
        self.assertEquals(result[1], hex_encode_as_bytes("chemfptest") + b"\t9425009")

    def test_daylight2(self):
        result = run_fps("--daylight --fp-tag daylight2", 2)
        self.assertEquals(result[0], hex_encode_as_bytes("Okie dokie pokie!") + b"\t9425004")
        self.assertEquals(result[1], hex_encode_as_bytes("Testing   1, 2, 3") + b"\t9425009")

    def test_daylight3(self):
        result = run_fps("--daylight --fp-tag daylight3", 2)
        self.assertEquals(result[0], b"\t9425004")
        self.assertEquals(result[1], b"\t9425009")

    def test_daylight_end_error(self):
        errmsg = run_exit("--daylight --fp-tag daylight-end-illegal", DECODER_SDF)
        self.assertIn("Could not daylight decode tag 'daylight-end-illegal' value '1P!_P'", errmsg)
        self.assertIn("Last character of encoding must be 1, 2, or 3, not 'P'", errmsg)
        self.assertIn("line 1", errmsg)
        self.assertIn("decoder.sdf", errmsg)

    def test_daylight_symbol_error(self):
        errmsg = run_exit("--daylight --fp-tag daylight-illegal", DECODER_SDF)
        self.assertIn("Could not daylight decode tag 'daylight-illegal' value '1P!_3'", errmsg)
        self.assertIn("Unknown encoding symbol", errmsg)
        self.assertIn("line 1", errmsg)
        self.assertIn("decoder.sdf", errmsg)

    def test_daylight_length_error(self):
        errmsg = run_exit("--daylight --fp-tag PUBCHEM_CACTVS_SUBSKEYS", DECODER_SDF)
        self.assertIn("Could not daylight decode tag 'PUBCHEM_CACTVS_SUBSKEYS' value 'AAADceB7sQ", errmsg)
        self.assertIn("Daylight binary encoding is of the wrong length", errmsg)
        self.assertIn("line 1", errmsg)
        self.assertIn("decoder.sdf", errmsg)


    def test_bad_decoding(self):
        _, _, msg = run_split_capture("--base64 --fp-tag binary17 --errors report")
        self.assertIn("Missing fingerprint tag 'binary17'", msg)
        self.assertIn("Skipping.", msg)

class TestBitSizes(unittest2.TestCase):
    def test_exact_fingerprint_bits(self):
        result = run("--binary --fp-tag binary3", DECODER_SDF)
        self.assertIn(b"#num_bits=3", result)
        
    def test_user_bits_match_fingerprint_bits(self):
        result = run("--binary --fp-tag binary3 --num-bits 3", DECODER_SDF)
        self.assertIn(b"#num_bits=3", result)
        self.assertIn(b"04\t9425004", result)
        self.assertIn(b"03\t9425009", result)

    def test_user_bits_disagree_with_fingerprint_bits(self):
        errmsg = run_exit("--binary --fp-tag binary3 --num-bits 2", DECODER_SDF)
        self.assertIn("has 3 bits", errmsg)
        self.assertIn(" 2", errmsg)

    def test_implied_from_fingerprint_bytes(self):
        result = run("--hex --fp-tag hex2", DECODER_SDF)
        self.assertIn(b"#num_bits=8", result)

    def test_user_bits_matches_fingerprint_bytes(self):
        result = run("--hex --fp-tag hex2 --num-bits 8", DECODER_SDF)
        self.assertIn(b"#num_bits=8", result)

    def test_user_bits_too_large_for_bytes(self):
        result = run_exit("--hex --fp-tag hex2 --num-bits 9", DECODER_SDF)
        self.assertIn("1 <= num-bits <= 8, not 9", result)

    def test_user_bits_acceptably_smaller_than_bytes(self):
        result = run("--hex --fp-tag hex2 --num-bits 6", DECODER_SDF)
        self.assertIn(b"#num_bits=6", result)

    def test_user_bits_too_much_smaller_than_bytes(self):
        result = run_exit("--hex --fp-tag hex16 --num-bits 56", DECODER_SDF)
        self.assertIn("57 <= num-bits <= 64, not 56", result)

class TestTitleProcessing(unittest2.TestCase):
    def test_title_from_title_tag(self):
        result = run("--hex --fp-tag hex2 --id-tag binary3", DECODER_SDF)
        self.assertIn(b"ab\t001", result)

    def test_missing_title_from_title_line(self):
        _, _, warning = run_split_capture("--hex --fp-tag hex2 --id-tag FAKE_TITLE --errors report")
        self.assertIn("Missing id tag 'FAKE_TITLE'", warning)
        self.assertIn("line 191", warning)
        self.assertIn("record #2", warning)
        self.assertIn("first line is '9425009'", warning)
        self.assertIn("Skipping.", warning)

    def test_missing_all_titles(self):
        _, _, warning = run_split_capture("--hex --fp-tag hex2 --id-tag DOES_NOT_EXIST --errors report")
        self.assertIn("Missing id tag 'DOES_NOT_EXIST'", warning)
        self.assertIn("line 1", warning)
        self.assertIn("record #1", warning)
        self.assertIn("line 191", warning)
        self.assertIn("record #2", warning)

    def test_control_return_in_title(self):
        filename = support.get_tmpfile(self, "control_return.sdf")
        with open(filename, "wb") as f:
            f.write(PUBCHEM_SDF_BYTES.replace(b"94250", b"QZ\r350"))
        header, output, warning = run_split_capture("--pubchem", source=filename)
        self.assertEqual(len(output), 19)
        for line in output:
            self.assertIn(b"QZ350", line)

    def test_nul_in_title(self):
        filename = support.get_tmpfile(self, "nul.sdf")
        with open(filename, "wb") as f:
            f.write(PUBCHEM_SDF_BYTES.replace(b"94250", b"QZ\x00350"))
        header, output, warning = run_split_capture("--pubchem", source=filename)
        self.assertEqual(len(output), 19)
        for line in output:
            self.assertIn(b"QZ350", line)
            
    def test_tab_in_title(self):
        filename = support.get_tmpfile(self, "tab.sdf")
        with open(filename, "wb") as f:
            f.write(PUBCHEM_SDF_BYTES.replace(b"94250", b"QZ\t350"))
        header, output, warning = run_split_capture("--pubchem", source=filename)
        self.assertEqual(len(output), 19)
        for line in output:
            self.assertIn(b"QZ350", line)

    def test_strip_title(self):
        filename = support.get_tmpfile(self, "strip_title.sdf")
        with open(filename, "wb") as f:
            f.write(PUBCHEM_SDF_BYTES.replace(b"9425009", b"\r\t\0"))
        header, output, warning = run_split_capture("--pubchem --errors report", source=filename)
        self.assertEqual(len(output), 18)
        for line in output:
            self.assertIn(b"9425", line)
        self.assertIn("Empty title in SD record after cleanup", warning)
        self.assertIn("record #2", warning)
        self.assertIn(r"'\r\t\x00'", warning)

class TestFingerprintProcessing(unittest2.TestCase):
    def test_fingerprint_length_changes(self):
        content = DECODER_SDF_BYTES.replace(b"0000000001011111111011011100101000000000", b"000000000101111111101101")
        with support.wrap_stdin(content):
            errmsg = run_exit("--binary --fp-tag binary40", source=None)
        self.assertIn("ERROR: Tag 'binary40' value '000000000101111111101101' has 24 bits but expected 40, "
                      "file '<stdin>', line 160, record #2: first line is '9425009'",
                      errmsg)

    def test_fingerprint_length_changes_ignore(self):
        content = DECODER_SDF_BYTES.replace(b"0000000001011111111011011100101000000000", b"000000000101111111101101")
        with support.wrap_stdin(content):
             header, lines, errmsg = run_split_capture("--binary --fp-tag binary40 --errors ignore", source=None)
        self.assertEqual(len(lines), 1)
        self.assertFalse(errmsg)

    def test_fingerprint_length_changes_report(self):
        content = DECODER_SDF_BYTES.replace(b"0000000001011111111011011100101000000000", b"000000000101111111101101")
        with support.wrap_stdin(content + DECODER_SDF_BYTES):
             header, lines, errmsg = run_split_capture("--binary --fp-tag binary40 --errors report", source=None)
        self.assertEqual(len(lines), 3)
        self.assertIn("ERROR: Tag 'binary40' value '000000000101111111101101' has 24 bits but expected 40, "
                      "file '<stdin>', line 160, record #2: first line is '9425009'",
                      errmsg)
    

        
FPS_9425004 = b"07de8d002000000000000000000000000080060000000c000000000000000080030000f8401800000030508379344c014956000055c0a44e2a0049200084e140581f041d661b10064483cb0f2925100619001393e10001007000000000008000000000000000400000000000000000\t9425004"

class TestParameters(unittest2.TestCase):
    def _check_valid(self, result, count=1):
        self.assertIn(b"#num_bits=881", result)
        self.assertIn(b"#software=CACTVS/unknown", result)
        self.assertIn(b"#type=CACTVS-E_SCREEN/1.0 extended=2", result)
        self.assertIn(FPS_9425004, result)
        self.assertIn(b"07de0d000000000000000000000000000080460300000c0000000000000000800f0000780038000000301083f920cc09695e0800d5c0e44e6e00492190844145dc1f841d261911164d039b8f29251026b9401313e0ec01007000000000000000000000000000000000000000000000\t9425009", result)
        if count != 1:
            self.assertEqual(result.count(FPS_9425004), count, result)

    def test_pubchem(self):
        result = run("--pubchem")
        self._check_valid(result)
        
    def test_in_sdf(self):
        result = run("--pubchem --in sdf", source=PUBCHEM_SDF)
        self._check_valid(result)
        
    def test_in_sdf_but_not_sdf(self):
        result = run_exit("--pubchem --in sdf", source=PUBCHEM_SDF_GZ)
        self.assertIn("Could not find a valid SD record", result)
        
    def test_in_sdf_gz(self):
        result = run("--pubchem --in sdf.gz", source=PUBCHEM_SDF_GZ)
        self._check_valid(result)

    def test_in_sdf_gz_but_not_sdf_gz(self):
        result = run_exit("--pubchem --in sdf.gz", source=PUBCHEM_SDF)
        self.assertIn("ERROR: Not a gzipped file", result)
        
    def test_multiple_input_files_with_mixed_compression(self):
        result = run(["--pubchem", PUBCHEM_SDF], source=PUBCHEM_SDF_GZ)
        self._check_valid(result, 2)

    def test_filename_does_not_exist(self):
        result = run_exit(["--pubchem"], source="/this/file/does/not/exist.sdf")
        self.assertIn("Structure file '/this/file/does/not/exist.sdf' does not exist", result)

    def test_filename_does_not_exist_with_multiple_files(self):
        result = run_exit(["--pubchem", PUBCHEM_SDF], source="/this/file/does/not/exist.sdf")
        self.assertIn("Structure file '/this/file/does/not/exist.sdf' does not exist", result)

    def test_from_stdin(self):
        with support.wrap_stdin(PUBCHEM_SDF_BYTES):
            result = run("--pubchem", source=None)
            self._check_valid(result)

    def test_wrong_encoding(self):
        result = run_exit("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --hex")
        self.assertIn("Could not hex decode tag 'PUBCHEM_CACTVS_SUBSKEYS' value", result)
        self.assertIn("Non-hexadecimal digit found", result)
        
    def test_wrong_encoding_ignore_errors(self):
        header, output, warning = run_split_capture("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --hex --errors ignore")
        self.assertEqual(len(output), 0)
        self.assertFalse(warning)
        
    def test_wrong_encoding_report_errors(self):
        header, output, warning = run_split_capture("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --hex --errors report")
        self.assertEqual(len(output), 0)
        self.assertEqual(warning.count("Could not hex decode tag"), 19)
        self.assertIn("'9425004'", warning)
        self.assertIn("'9425009'", warning)
        self.assertIn("'9425046'", warning)

    def test_bitlength_changes(self):
        content = PUBCHEM_SDF_BYTES.replace(
            b"AAADceB7sABAAAAAAAAAAAAAAAAAAWLAAAAwAAAAAAAAAFgB/AAAHgQYAAAACAjB1gQywbJqEAiuASVyVACT9KBhij"
            b"pa+D24ZJgIYLLg0fGUpAhgmADoyAcYCAAAAAAAAAAAAQAAAAAAAAACAAAAAAAAAA==", b"QW5kcmV3IERhbGtl")
        with support.wrap_stdin(content):
            result = run_exit("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --base64", source=None)
        self.assertIn("ERROR: Tag 'PUBCHEM_CACTVS_SUBSKEYS' value", result)
        self.assertIn("'QW5kcmV3IERhbGtl'", result)
        self.assertIn("has 12 bytes but expected 115", result)
        self.assertIn("record #18", result)
        
    def test_bitlength_changes_ignore_errors(self):
        content = PUBCHEM_SDF_BYTES.replace(
            b"AAADceB7sABAAAAAAAAAAAAAAAAAAWLAAAAwAAAAAAAAAFgB/AAAHgQYAAAACAjB1gQywbJqEAiuASVyVACT9KBhij"
            b"pa+D24ZJgIYLLg0fGUpAhgmADoyAcYCAAAAAAAAAAAAQAAAAAAAAACAAAAAAAAAA==", b"QW5kcmV3IERhbGtl")
        with support.wrap_stdin(content):
            header, output, warning = run_split_capture("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --base64 --errors ignore", source=None)
        self.assertEqual(len(output), 18)
        self.assertFalse(warning)
        
    def test_bitlength_changes_report_errors(self):
        content = PUBCHEM_SDF_BYTES.replace(
            b"AAADceB7sABAAAAAAAAAAAAAAAAAAWLAAAAwAAAAAAAAAFgB/AAAHgQYAAAACAjB1gQywbJqEAiuASVyVACT9KBhij"
            b"pa+D24ZJgIYLLg0fGUpAhgmADoyAcYCAAAAAAAAAAAAQAAAAAAAAACAAAAAAAAAA==", b"QW5kcmV3IERhbGtl")
        with support.wrap_stdin(content):
            header, output, warning = run_split_capture("--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --base64 --errors report", source=None)
        self.assertEqual(len(output), 18)
        self.assertIn("ERROR: Tag 'PUBCHEM_CACTVS_SUBSKEYS' value", warning)
        self.assertIn("'QW5kcmV3IERhbGtl'", warning)
        self.assertIn("has 12 bytes but expected 115", warning)
        self.assertIn("record #18", warning)
        
    def test_output_file_does_not_exist(self):
        result = run_exit("--pubchem -o /this/also/does/not/exist/at/all.fps")
        self.assertIn("Cannot open output file", result)
        self.assertIn("'/this/also/does/not/exist/at/all.fps'", result)
        
    def test_output_file_uses_unsupported_format(self):
        result = run_exit("--pubchem -o /this/also/does/not/exist/at/all.fps.xz")
        self.assertIn("Cannot open output fingerprint file", result)
        self.assertIn("chemfp does not yet support xz compression", result)

        
class TestBadArgs(unittest2.TestCase):
    def test_missing_fp_tag(self):
        msg = run_exit("")
        self.assertIn("argument --fp-tag is required", msg)

    def test_num_bits_positive(self):
        msg = run_exit("--fp-tag SPAM --num-bits 0")
        self.assertIn("--num-bits must be a positive integer", msg)
        msg = run_exit("--fp-tag SPAM --num-bits -1")
        self.assertIn("--num-bits must be a positive integer", msg)

    def test_bad_char(self):
        msg = run_exit("--fp-tag SPAM --software this\bthat")
        self.assertIn("--software", msg)
        self.assertIn("'\\x08'", msg)

@unittest2.skipUnless(support.has_chemfp_converters, "chemfp_converters is not installed")
class TestFlushSupport(unittest2.TestCase):
    def _check_reader(self, reader):
        self.assertEqual(reader.metadata.num_bits, 128)
        rows = list(reader)
        self.assertEqual(rows, [
            (u"9425004", b"\x47\x72\x65\x65\x74\x69\x6e\x67\x73\x2c\x20\x68\x75\x6d\x61\x6e"),
            (u"9425009", b"\x62\x6c\x61\x68\x62\x6c\x61\x68\x73\x70\x61\x6d\x62\x6c\x61\x68"),
            ])
        
    def test_flush_output(self):
        from chemfp_converters import flush
        filename = support.get_tmpfile(self, "output.flush")
        msg = run(["--fp-tag", "base64_16", "--base64", "-o", filename], DECODER_SDF)
        self.assertFalse(msg)
        
        with flush.open_flush(filename) as reader:
            self._check_reader(reader)
            
    def test_flush_output_format(self):
        from chemfp_converters import flush
        filename = support.get_tmpfile(self, "output.xyz")
        msg = run(["--fp-tag", "base64_16", "--base64", "-o", filename, "--out", "flush"], DECODER_SDF)
        self.assertFalse(msg)
        
        with flush.open_flush(filename) as reader:
            self._check_reader(reader)

    def test_flush_stdout(self):
        from chemfp_converters import flush
        filename = support.get_tmpfile(self, "output.flush")
        real_stdout = sys.stdout
        bytes_stdout = io.BytesIO()
        sys.stdout = bytes_stdout
        try:
            try:
                sdf2fps.main([DECODER_SDF, "--fp-tag", "base64_16", "--base64", "--out", "flush"])
            except SystemExit as err:
                raise AssertionError("Why did I get a SystemExit?: %s" % (err,))
        finally:
            sys.stdout = real_stdout
        output = bytes_stdout.getvalue()
        with open(filename, "wb") as f:
            f.write(output)
        with flush.open_flush(filename) as reader:
            self._check_reader(reader)
        

    def test_flush_wrong_size_stdout(self):
        msg = run_exit(["--pubchem", "--out", "flush"])
        # Not the best of error messages, but it will do for now.
        self.assertIn("metadata num_bytes (111) must be a multiple of 4", msg)
            
    def test_flush_wrong_size_filename(self):
        filename = support.get_tmpfile(self, "output.flush")
        msg = run_exit(["--pubchem", "-o", filename])
        # Not the best of error messages, but it will do for now.
        self.assertIn("metadata num_bytes (111) must be a multiple of 4", msg)
            
if __name__ == "__main__":
    unittest2.main()



# Test the fingerprint reordering implementation

# Note: this is the ordering by popcount, and NOT the ordering of the results!

from __future__ import absolute_import, with_statement
import unittest2
from cStringIO import StringIO
import struct

import chemfp
from chemfp.bitops import byte_popcount

def _load(fingerprints, reorder):
    if len(fingerprints) == 0:
        num_bits = 16
    else:
        num_bits = len(fingerprints[0])*8
    id_fps = ((str(i), fp) for (i, fp) in enumerate(fingerprints))
    return chemfp.load_fingerprints(id_fps,
                                    metadata=chemfp.Metadata(num_bits=num_bits),
                                    reorder=reorder, alignment=1)

def verify_popcount_indices(arena):
    assert len(arena.popcount_indices) % 4 == 0
    format = "i"*(len(arena.popcount_indices)//4)
    values = struct.unpack(format, arena.popcount_indices)
    assert len(values) == arena.metadata.num_bits+2, (len(values), arena.metadata.num_bits)
    assert values[-1] == len(arena), (values[-1], len(arena), values)
    for i in range(len(values)-1):
        start = values[i]
        end = values[i+1]

        for j in range(start, end):
            fp = arena[j][1]
            assert byte_popcount(fp) == i, (byte_popcount(fp), i, start, end)
    
    

class TestReorder(unittest2.TestCase):
    def test_empty(self):
        arena = _load([], True)
        self.assertEquals(arena.arena, "")
        verify_popcount_indices(arena)
        
        arena = _load([], False)
        self.assertEquals(arena.arena, "")
        self.assertEquals(arena.popcount_indices, "")

    def test_single(self):
        arena = _load(["1234"], True)
        self.assertEquals(arena.arena, "1234")
        verify_popcount_indices(arena)
        
        arena = _load(["1234"], False)
        self.assertEquals(arena.arena, "1234")

    def test_two(self):
        arena = _load(["AA", "CC"], True)
        self.assertEquals(arena.arena, "AACC")
        verify_popcount_indices(arena)
        
        arena = _load(["AA", "CC"], False)
        self.assertEquals(arena.arena, "AACC")
        
        arena = _load(["CC", "AA"], True)
        self.assertEquals(arena.arena, "AACC")
        verify_popcount_indices(arena)
        
        arena = _load(["CC", "AA"], False)
        self.assertEquals(arena.arena, "CCAA")

    def test_every_bit_unsorted(self):
        arena = _load([chr(i) for i in range(256)], False)
        self.assertEquals(arena.arena, "".join(chr(i) for i in range(256)))

    def test_every_bit_sorted(self):
        # This is a bit tricker since there's no guaranteed order
        # of the contents of the arena
        
        all_bytes = [chr(i) for i in range(256)]
        expected = sorted(byte_popcount(fp) for fp in all_bytes)

        arena = _load(all_bytes, True)
        popcounts = map(byte_popcount, arena.arena)
        self.assertEquals(popcounts, expected)
        verify_popcount_indices(arena)

        arena = _load(all_bytes[::-1], True)
        popcounts = map(byte_popcount, arena.arena)
        self.assertEquals(popcounts, expected)
        verify_popcount_indices(arena)

    def test_every_bit_sorted_tripled(self):
        # This is a bit tricker since there's no guaranteed order
        # of the contents of the arena
        
        all_bytes = [chr(i) for i in range(256)]
        all_bytes *= 3
        expected = sorted(byte_popcount(fp) for fp in all_bytes)

        arena = _load(all_bytes, True)
        popcounts = map(byte_popcount, arena.arena)
        self.assertEquals(popcounts, expected)
        verify_popcount_indices(arena)

        arena = _load(all_bytes[::-1], True)
        popcounts = map(byte_popcount, arena.arena)
        self.assertEquals(popcounts, expected)
        verify_popcount_indices(arena)


    def test_all_ones(self):
        arena = _load([chr(255), chr(255), chr(255), chr(255)], True)
        self.assertEquals(arena.arena, "\xff\xff\xff\xff")
        verify_popcount_indices(arena)

if __name__ == "__main__":
    unittest2.main()



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import with_statement, print_function

import sys

from .. import ChemFPError, open_fingerprint_writer
from .. import argparse, io, rdkit, types
from . import cmdsupport

########### Configure the command-line parser

epilog = ""

if rdkit.PATTERN_VERSION is None:
    epilog += """\
    
This version of RDKit does not support the 'Pattern' fingerprints.
"""

if not rdkit.HAS_AVALON:
    epilog += """\
    
This version of RDKit does not support the Avalon fingerprints. To use them,
reinstall RDKit with pyAvalonTools.
"""

    
epilog = """\

This program guesses the input structure format based on the filename
extension. If the data comes from stdin, or the extension name us
unknown, then use "--in" to change the default input format. The
supported format extensions are:

  File Type      Valid FORMATs (use gz if compressed)
  ---------      ------------------------------------
   SMILES        smi, ism, usm, can, smi.gz, ism.gz, usm.gz, can.gz
   SDF           sdf, mol, sd, mdl, sdf.gz, mol.gz, sd.gz, mdl.gz
"""

parser = argparse.ArgumentParser(
    description="Generate FPS fingerprints from a structure file using RDKit",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    conflict_handler="resolve",
    )

_base = rdkit._base_2048

# --RDK and --morgan both have fpSize but argparse doesn't allow the
# same option in different groups. Especially with different defaults.

rdkit._base_fpsize.add_argument_to_argparse("fpSize", parser)

rdk_group = parser.add_argument_group("RDKit topological fingerprints")
rdk_group.add_argument("--RDK", action="store_true",
                       help="generate RDK fingerprints (default)")
_base.add_argument_to_argparse("minPath", rdk_group)
_base.add_argument_to_argparse("maxPath", rdk_group)
_base.add_argument_to_argparse("nBitsPerHash", rdk_group)
_base.add_argument_to_argparse("useHs", rdk_group)


morgan_group = parser.add_argument_group("RDKit Morgan fingerprints")

morgan_group.add_argument("--morgan", action="store_true",
                          help="generate Morgan fingerprints")

_morgan = rdkit.RDKitMorganFingerprintFamily_v1
_morgan.add_argument_to_argparse("radius", morgan_group)
_morgan.add_argument_to_argparse("useFeatures", morgan_group)
_morgan.add_argument_to_argparse("useChirality", morgan_group)
_morgan.add_argument_to_argparse("useBondTypes", morgan_group)

torsion_group = parser.add_argument_group("RDKit Topological Torsion fingerprints")
torsion_group.add_argument("--torsions", action="store_true",
                           help="generate Topological Torsion fingerprints")
rdkit.RDKitTorsionFingerprintFamily_v1.add_argument_to_argparse(
    "targetSize", torsion_group)

pair_group = parser.add_argument_group("RDKit Atom Pair fingerprints")
pair_group.add_argument("--pairs", action="store_true",
                        help="generate Atom Pair fingerprints")
rdkit.RDKitTorsionFingerprintFamily_v1.add_argument_to_argparse(
    "minLength", pair_group)
rdkit.RDKitTorsionFingerprintFamily_v1.add_argument_to_argparse(
    "maxLength", pair_group)



maccs_group = parser.add_argument_group("166 bit MACCS substructure keys")
maccs_group.add_argument(
    "--maccs166", action="store_true", help="generate MACCS fingerprints")

avalon_help = "generate Avalon fingerprints"
if rdkit.HAS_AVALON:
    has_avalon = True
else:
    has_avalon = False
    avalon_help += " (not available for this RDKit install)"
avalon_group = parser.add_argument_group("Avalon fingerprints")
avalon_group.add_argument(
    "--avalon", action="store_true", help=avalon_help)
rdkit._base_avalon.add_argument_to_argparse("isQuery", avalon_group)
rdkit._base_avalon.add_argument_to_argparse("bitFlags", avalon_group)

pattern_group = parser.add_argument_group("RDKit Pattern fingerprints")
pattern_group.add_argument("--pattern", action="store_true",
                           help="generate (substructure) pattern fingerprints")


substruct_group = parser.add_argument_group("881 bit substructure keys")
substruct_group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints")


rdmaccs_group = parser.add_argument_group("ChemFP version of the 166 bit RDKit/MACCS keys")
rdmaccs_group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="generate 166 bit RDKit/MACCS fingerprints (version 2)")
rdmaccs_group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")


def fromAtom_list(s):
    t = s.strip()
    if not t:
        raise argparse.ArgumentTypeError(
            "must contain a comma-separated list of atom indices")

    terms = t.split(",")
    seen = set()
    for term in terms:
        if not term.isdigit():
            raise argparse.ArgumentTypeError(
                "term %r must be a non-negative integer"
                % (term,))
        seen.add(int(term))
    return sorted(seen)
        
## from_group = parser.add_mutually_exclusive_group()
## from_group.add_argument("--all-atoms", action="store_true")
## from_group.add_argument("--unique-atoms", action="store_true")
parser.add_argument("--from-atoms", dest="fromAtoms", metavar="INT,INT,...",
                    type=fromAtom_list, default=None,
                    help="fingerprint generation must use these atom indices (out of range indices are ignored)")


parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id (SD files only)")
    
parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default guesses from filename)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")

parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_version(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")

#########

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)
    cmdsupport.mutual_exclusion(parser, args, "RDK",
                                ("maccs166", "RDK", "substruct", "rdmaccs", "rdmaccs/1", "morgan",
                                 "torsions", "pairs", "pattern", "avalon"))
    # If size isn't specified, use the appropriate size that fingerprint
    if args.fpSize is None:
        if args.avalon:
            args.fpSize = rdkit.AVALON_NBITS
        else:
            args.fpSize = rdkit.NUM_BITS

    if args.maccs166:
        opener = types.get_fingerprint_family("RDKit-MACCS166")()
    elif args.RDK:
        fpSize = args.fpSize or rdkit.NUM_BITS
        minPath = args.minPath
        maxPath = args.maxPath
        nBitsPerHash = args.nBitsPerHash
        if maxPath < minPath:
            parser.error("--minPath must not be greater than --maxPath")

        useHs = args.useHs

        opener = types.get_fingerprint_family("RDKit-Fingerprint")(
            minPath=minPath,
            maxPath=maxPath,
            fpSize=fpSize,
            nBitsPerHash=nBitsPerHash,
            useHs=useHs,
            fromAtoms=args.fromAtoms)

    elif args.substruct:
        opener = types.get_fingerprint_family("ChemFP-Substruct-RDKit")()
    elif args.rdmaccs:
        opener = types.get_fingerprint_family("RDMACCS-RDKit")()
    elif getattr(args, "rdmaccs/1"):
        opener = types.get_fingerprint_family("RDMACCS-RDKit/1")()
    elif args.morgan:
        opener = types.get_fingerprint_family("RDKit-Morgan")(
            radius=args.radius,
            fpSize=args.fpSize,
            useFeatures=args.useFeatures,
            useChirality=args.useChirality,
            useBondTypes=args.useBondTypes,
            fromAtoms=args.fromAtoms)

    elif args.torsions:
        opener = types.get_fingerprint_family("RDKit-Torsion")(
            fpSize=args.fpSize,
            targetSize=args.targetSize,
            fromAtoms=args.fromAtoms)
    elif args.pairs:
        minLength = args.minLength
        maxLength = args.maxLength
        if maxLength < minLength:
            parser.error("--minLength must not be greater than --maxLength")
        opener = types.get_fingerprint_family("RDKit-AtomPair")(
            fpSize=args.fpSize,
            minLength=minLength,
            maxLength=maxLength,
            fromAtoms=args.fromAtoms)
    elif args.pattern:
        opener = types.get_fingerprint_family("RDKit-Pattern")(
            fpSize=args.fpSize)
    elif args.avalon:
        if rdkit.HAS_AVALON:
            opener = types.get_fingerprint_family("RDKit-Avalon")(
                fpSize=args.fpSize,
                isQuery=args.isQuery,
                bitFlags=args.bitFlags)
        else:
            parser.error("Cannot generate --avalon fingerprints. The RDKit toolkit was not compiled with Avalon support.")
        
    else:
        raise AssertionError("Unknown fingerprinter")

    if not rdkit.is_valid_format(args.format):
        parser.error("Unsupported format specifier: %r" % (args.format,))

    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")

    cmdsupport.check_filenames(parser, None, args.format, args.filenames)

    # Ready the input reader/iterator
    metadata, reader = cmdsupport.read_multifile_structure_fingerprints(
        opener, args.filenames, format = args.format, id_tag = args.id_tag, 
        reader_args = {}, errors = args.errors)

    location = io.Location.from_destination(args.output)
    try:
        writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
    except (IOError, ValueError) as err:
        msg = "Cannot open output fingerprint file: %s" % (err,)
        if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
            msg += ".\nUse --out to specify 'fps' or 'fps.gz'."
        parser.error(msg)
        
    try:
        with writer:
            writer.write_fingerprints(reader)
    except (ChemFPError, IOError, EOFError) as err:
        sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
        raise SystemExit(1)

if __name__ == "__main__":
    main()




from __future__ import with_statement

import unittest2
from chemfp import types

try:
    from chemfp import openeye
except ImportError:
    skip_openeye = True
else:
    skip_openeye = False

try:
    from chemfp import rdkit
except ImportError:
    skip_rdkit = True
else:
    skip_rdkit = False


# These tests are to improve overall test coverage

class ParseType(unittest2.TestCase):
    def test_no_type_string(self):
        with self.assertRaisesRegexp(ValueError, r"Empty fingerprint type \(''\)"):
            types.parse_type("")
        with self.assertRaisesRegexp(ValueError, r"Empty fingerprint type \(' '\)"):
            types.parse_type(" ")

    def test_unknown_fingerprint_family(self):
        with self.assertRaisesRegexp(ValueError, "Unknown fingerprint family 'Blah-Blah/1'"):
            types.parse_type("Blah-Blah/1")

class TestOEFingerprintTypes(unittest2.TestCase):
    def test_missing_equals(self):
        with self.assertRaisesRegexp(ValueError, "Term 'atype' in type 'OpenEye-Path atype' must have one and only one '='"):
            types.parse_type("OpenEye-Path atype")

    def test_extra_equals(self):
        with self.assertRaisesRegexp(ValueError, "Term 'atype=DefaultAtom=DefaultAtom' in type 'OpenEye-Path atype=DefaultAtom=DefaultAtom' must have one and only one '='"):
            types.parse_type("OpenEye-Path atype=DefaultAtom=DefaultAtom")

    def test_duplicate_names(self):
        with self.assertRaisesRegexp(ValueError, "Duplicate name 'atype' in type 'OpenEye-Path atype=DefaultAtom atype=InRing'"):
            types.parse_type("OpenEye-Path atype=DefaultAtom atype=InRing")

    def test_unknown_name(self):
        with self.assertRaisesRegexp(ValueError, "Unknown name 'atomtype' in type 'OpenEye-Path atomtype=DefaultAtom'"):
            types.parse_type("OpenEye-Path atomtype=DefaultAtom")

    def test_negative_minbonds(self):
        with self.assertRaisesRegexp(ValueError, "Unable to parse minbonds value '-1' in type 'OpenEye-Path minbonds=-1'"):
            types.parse_type("OpenEye-Path minbonds=-1")

    def test_make_fingerprinter_from_type_with_empty_string(self):
        family = types.get_fingerprint_family("OpenEye-Path")
        with self.assertRaisesRegexp(ValueError, r"Empty fingerprint type \(''\)"):
            family.make_fingerprinter_from_type("")

    def test_compare_types_for_equality(self):
        t1 = types.parse_type("OpenEye-Path")
        self.assertEquals(t1, t1)
        t2 = types.parse_type("OpenEye-Path")
        self.assertEquals(t1, t2)

    def test_compare_types_for_inequality(self):
        t1 = types.parse_type("OpenEye-Path")
        t2 = types.parse_type("OpenEye-Path atype=InRing")
        self.assertNotEquals(t1, t2)
        
TestOEFingerprintTypes = unittest2.skipIf(skip_openeye, "OEChem not installed")(TestOEFingerprintTypes)

class TestRDKitFingerprintTypes(unittest2.TestCase):
    def test_passing_0_to_positive_int(self):
        with self.assertRaisesRegexp(ValueError, "Unable to parse fpSize value '0' in type 'RDKit-Fingerprint fpSize=0'"):
            types.parse_type("RDKit-Fingerprint fpSize=0")
            
    def test_passing_negative_to_nonnegative_int(self):
        with self.assertRaisesRegexp(ValueError, r"Unable to parse fpSize value '\+2' in type 'RDKit-Fingerprint fpSize=\+2'"):
            types.parse_type("RDKit-Fingerprint fpSize=+2")

    def test_passing_something_other_than_0_or_1(self):
        with self.assertRaisesRegexp(ValueError, r"Unable to parse useHs value '2' in type 'RDKit-Fingerprint useHs=2'"):
            types.parse_type("RDKit-Fingerprint useHs=2")

TestRDKitFingerprintTypes = unittest2.skipIf(skip_rdkit, "OEChem not installed")(TestRDKitFingerprintTypes)

if __name__ == "__main__":
    unittest2.main()
